'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the size based variables for this application
' (C_Prefix indicates constants).
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCDSize(LCD_LineLength, LCD_Lines)

    C_LineLength = LCD_LineLength
    C_Lines = LCD_Lines
    '
    C_LineStart(1) = 0        ' Start address for line 1
    C_LineStart(2) = 64       ' Start address for line 2
    C_LineStart(3) = 20       ' Start address for line 3
    C_LineStart(4) = 84       ' Start address for line 4

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Change the Maximite / DM pins allocate for LCD control and data.
' argument #1 is the pin to use for the LCD RS line
' argument #2 is the pin to use for the LCD EN line
' argument #3 is the pin to use for the LCD D1 line
' argument #4 is the pin to use for the LCD D2 line
' argument #5 is the pin to use for the LCD D3 line
' argument #6 is the pin to use for the LCD D4 line
' argument #7 is the pin to use for the LCD RW line (optional)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB ChangeLCDPins(_RS, _EN, _D1, _D2, _D3, _D4, _RW)
    RS = _RS
    EN = _EN
    D1 = _D1
    D2 = _D2
    D3 = _D3
    D4 = _D4
    IF _RW <> 0 Then
        IF RWenabled = 1 THEN   ' true
            RW = _RW
        ELSE
            ? "You must allow use of the Read/Write line by calling sub LCDenableRW"
        ENDIF
    ENDIF
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Initialise the LCD hardware
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB InitLCD
    SETPIN RS, 9                                ' open collector
    SETPIN EN, 9                                ' open collector
    SETPIN D1, 9                                ' open collector
    SETPIN D2, 9                                ' open collector
    SETPIN D3, 9                                ' open collector
    SETPIN D4, 9                                ' open collector
    IF RWenabled THEN SETPIN RW, 2              ' digital input (optional)

    LCD_Nibble_Write &B0011, 0, 5               ' reset
    LCD_Nibble_Write &B0011, 0, 5               ' reset
    LCD_Nibble_Write &B0011, 0, 5               ' reset
    LCD_Nibble_Write &B0010, 0, 2               ' 4 bit mode
    LCD_Nibble_Write &B0010 : LCD_Nibble_Write &B1100       ' 4 bits, 2 lines
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B1100       ' display on, no cursor
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B0110       ' increment on write
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B0001, 0, 2 ' clear the display, pause 2
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Display a line on the LCD
' argument #1 is the line to be used (1 to maximum line number)
' argument #2 is the line of characters to be displayed (any length)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB PrintLCD ( LineNumber, Line$ )
    LOCAL i, c, fillSpaces, fullLine$, controlByte

    controlByte = 128 + C_LineStart(LineNumber)

    ' first, send the cursor position (in two nibbles)
    LCD_WriteByte(controlByte)

    ' next, make sure the line is right padded with spaces, if needed
    fillSpaces = C_LineLength - LEN(Line$)
    IF fillSpaces > 0 THEN
        fullLine$ = Line$ + SPACE$(fillSpaces)
    ELSE
        fullLine$ = Line$
    ENDIF

    ' then send the text character by character (two nibbles per character)
    ' note - different to LCD.BAS. This allows text longer than line length.
    FOR i = 1 TO LEN(Line$)
        LCD_WriteByte ASC(MID$(fullLine$, i, 1)), 1
    NEXT i

END SUB


SUB LCD_WriteByte(_Byte, _Flag, _Wait_mSec)

    LCD_Nibble_Write INT(_Byte/16), _Flag, _Wait_mSec
    LCD_Nibble_Write _Byte, _Flag, _Wait_mSec

END SUB


SUB LCD_ReadByte(_Byte)
	' untested
    LOCAL b

    LCD_Nibble_Read(b)
    _Byte = b  * 16
    LCD_Nibble_Read(b)
    _Byte = _Byte OR b
    ? "Byte read is "; _Byte; " or, in binary, "; BIN$(_Byte)

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read the lower 4 bits (called a nibble) from the LCD
' argument #1 returns the nibble
' argument #2 is delay after the data has been received (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble_Read ( Data, Wait_mSec )
    LOCAL Value
    ' untested
    PIN(RS) = 1                 ' data
    PIN(EN) = 1 : PIN(EN) = 0
    IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
    ? "Bits read are: "; PIN(D1); " "; PIN(D2); " "; PIN(D3); " "; PIN(D4)
    Value = PIN(D1) ' low order bit
    Value = Value OR (PIN(D2) * 2)
    Value = Value OR (PIN(D3) * 4)
    Value = Value OR (PIN(D4) * 8)

    ? "Nibble value is "; Value; ", which is binary "; Bin$(Value)

    Data = Value

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Send the lower 4 bits (called a nibble) to the LCD
' argument #1 is the nibble to send
' argument #2 is true if data, false if command (default is command)
' argument #3 is delay after the data has been sent (default is zero)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCD_Nibble_Write ( Data, Flag, Wait_mSec )
    PIN(RS) = Flag
    PIN(D1) = Data AND &B00000001
    PIN(D2) = Data AND &B00000010
    PIN(D3) = Data AND &B00000100
    PIN(D4) = Data AND &B00001000
    PIN(EN) = 1 : PIN(EN) = 0
    IF Wait_mSec > 0 THEN PAUSE Wait_mSec       ' modified 30/1/2012. Restore to Wait_mSec when Pause 0 bug corrected.
END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Move cursor to a location on the LCD screen.
' argument #1 is the line to be used (1 to maximum line number)
' argument #2 is the column to be used (1 to line length)
' tested OK.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDLocate(Line, Column)
    LOCAL controlByte

    controlByte = 128 + C_LineStart(Line) + Column - 1
    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Scroll a message across the LCD screen
' argument #1 is the line to be used (1 to 4)
' argument #2 is the message to display
' tested ok.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

SUB ScrollLCDMessage(line, Message$)
    LOCAL MsgLen, Mask_Offset, i, c

    MsgLen = LEN(Message$)
    FOR Mask_Offset = 0 TO MsgLen - C_LineLength ' keep track of where we're starting the display window
        LCDLocate(line,1)                     ' go to the start of line
        FOR i = 1 TO C_LineLength
            c = ASC(MID$(Message$, Mask_Offset + i, 1))
            LCD_WriteByte c, 1
        NEXT i
        IF Mask_offset = 0 THEN
            PAUSE 1000  ' initial - so user has time to read the initial part of the message
        ELSE
            PAUSE 400   ' each char - so screen has time to refresh, and user can read it
        ENDIF
    NEXT Mask_Offset

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Enable the Read Write Flag.
' Most people won't use this, but it's here for completeness and
' so the Busy Flag and address can be read from the LCD
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDenableRW

    RWenabled = 1   ' set to true

END SUB

' Set LCD display, cursor, and cursor blink controls

SUB LCD_Display(Param)     ' C_On or C_Off

    LCDControl(Param, _LCD_Cursor, _LCD_Blink)

END SUB

SUB LCD_Cursor(Param)      ' C_On or C_Off

    LCDControl(_LCD_Display, Param, _LCD_Blink)

END SUB

SUB LCD_Blink(Param)       ' C_On or C_Off

    LCDControl(_LCD_Display, _LCD_Cursor, Param)

END SUB


'****************************************************************
' The following subroutines implement the LCD controls.
'****************************************************************

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Clear the LCD screen.
' Clears entire display and sets DDRAM address 0 in address counter
' tested ok
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDClearScreen
    LOCAL controlByte

    controlByte = &B00000001
    LCD_WriteByte controlByte
    pause 2

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set cursor to Home position.
' Sets DDRAM address 0 in address counter.  Also returns display
' from being shifted to original position. DDRAM contents remain
' unchanged.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDHome
    LOCAL controlByte

    controlByte = &B00000010
    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Sets cursor move direction and specifies display shift.
' These operations are performed during data write and read.
' argument #1 is the move direction (C_Increment or C_Decrement)
' argument #2 is the display shift (C_On or C_Off)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDEntryMode(MoveDir, Shift)
    LOCAL controlByte

    controlByte = &B00000100

    _LCD_MoveDir = MoveDir
    IF _LCD_MoveDir = C_Increment THEN
        controlByte = controlByte OR &B00000010
    ENDIF

    _LCD_Shift = Shift
    IF _LCD_Shift = C_On THEN
        controlByte = controlByte OR &B00000001
    ENDIF

    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set LCD display, cursor, and cursor blink controls
' argument #1 is the Display      status C_On or C_Off
' argument #2 is the Cursor       status C_On or C_Off
' argument #3 is the Cursor Blink status C_On or C_Off
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDControl(Display, Cursor, Blink)
    LOCAL controlByte

    controlByte = &B00001000

    _LCD_Display = Display
    IF _LCD_Display = C_On THEN
        controlByte = controlByte OR &B00000100
    ENDIF

    _LCD_Cursor = Cursor
    IF _LCD_Cursor = C_On THEN
        controlByte = controlByte OR &B00000010
    ENDIF

    _LCD_Blink = Blink
    IF _LCD_Blink = C_On THEN
        controlByte = controlByte OR &B00000001
    ENDIF

    LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Cursor and Display shift.
' Moves cursor and shifts display without changing DDRAM contents.
' argument #1 is the type (C_Display_Shift or C_Cursor_Move)
' argument #2 is the shift direction (C_Left or C_Right)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDShift(Type, Direction)
    LOCAL controlByte

    controlByte = &B00010000

    _LCD_Move_Type = Type
    IF _LCD_Move_Type = C_Display_Shift THEN
        controlByte = controlByte OR &B00001000
    ENDIF

    _LCD_Direction = Direction
    IF _LCD_Direction = C_Right THEN
        controlByte = controlByte OR &B00000100
    ENDIF

     LCD_WriteByte controlByte

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the interface data length, number of display lines and
' character font.
' argument #1 is the number of bits used by the interface (4 or 8)
' note: this  code is currently only set up to use 4 bits
' argument #2 is the number of lines (1 or 2)
' TODO: Update with how to select 4 lines
' argument #3 is the font size (C_Large or C_Normal)
' Large is 5 x 10 dots, Normal is 5 x 8 dots.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDFunctionSet(_Bits, _Lines, _Font)
    LOCAL controlByte

    controlByte = &B00100000

    _LCD_Bits = _Bits
    IF _Bits = 8 THEN
        ? "Warning: 8 bit mode is not currently supported!"
        ? "Ignoring instruction.  Using default 4 bit interface."
        ' controlByte = controlByte OR &B00010000
        EXIT SUB
    ENDIF

    _LCD_Lines = _Lines
    IF _Lines = 2 THEN   ' also used for 4 line displays
        controlByte = controlByte OR &B00001000
    ENDIF

    _LCD_Font = _Font
    IF _LCD_Font = C_Large THEN
    	IF _LCD_Lines <> 1 THEN
    		? "The Large font (5 x 10) only works with 1 line displays."
    		? "Ignoring instruction.  Using default Normal font."
    	ELSE
        	controlByte = controlByte OR &B00000100
        ENDIF
    ENDIF

    ' ? Bin$(controlByte)

    LCD_WriteByte controlByte

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' CGRAM data is sent and received after this setting.
' argument #1 is the CGRAM address (6 bits, so 0 to 63)
'
' In the character generator RAM, the user can rewrite character
' patterns by program.
' For 5 x  8 dots, eight character patterns can be written, and
' for 5 x 10 dots, four  character patterns can be written.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetCGRAMaddr(_Addr)
    LOCAL controlByte

    controlByte = &B01000000

    IF _Addr > 63 THEN
        ? "CGRAM Address "; _Addr; " is too high. It must be 0 to 63."
    ELSE
        controlByte = controlByte OR _Addr
        ' ? Bin$(controlByte)
        LCD_WriteByte controlByte
    ENDIF

END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Display Data RAM (DDRAM) address.
' DDRAM data is sent and received after this setting.
' argument #1 is the DDRAM address (7 bits, so 0 to 127)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDSetDDRAMaddr(_Addr)
    LOCAL controlByte

    controlByte = &B10000000

    IF _Addr > 127 THEN
        ? "DDRAM Address "; _Addr; " is too high (should be 0 to 127)"
    ELSE
        controlByte = controlByte OR _Addr
        ' ? Bin$(controlByte)
        LCD_WriteByte controlByte
    ENDIF

END SUB


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read busy flag.
' Returns the Busy Flag (BF) indicating internal operation is being
' performed and returns address counter contents.
' bf=1 - still internally operating. b=0 - instructions acceptable
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB LCDreadBusyFlag(_Status, _Addr)
    LOCAL myByte, BF

    LCD_ReadByte(myByte)
    BF    = myByte AND &B10000000  ' busy flag
    _Addr = myByte AND &B01111111  ' Address Counter

    _Status = (BF = 128)    'convert to a value of 0 or 1

END SUB

