' Todo - code for Read Status and address from LCD
' Todo - allow 8 data lines
' Todo - sub LCDHelp to display all subs and parameters
' Todo - establish naming conventions for parameters, subs
' Todo - doco.  Include reference to tutorials.
' Todo - test with all types of LCD
' Todo - add test to check version >= 3.1
' Bug: TestLCDFunctionSet - first test

'
' demonstration program
' send lines to the LCD
'
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' LCD driver for MMBasic 3.1 or later (uses defined subroutines)
' Geoff Graham, Jan 2012.
' Modified by James Deakins to include 20 x 4 LCDs, Feb/Mar 2012.
'
' Will drive standard 16 x 2 or 20 x 4 LCDs, and should drive all others
' For example:  futurlec.com       LCD16X2
'               altronics.com.au   Z7001
'               jaycar.com.au      QP5512
'                                  HD44780 (20 x 4)
'
' To use:
'- Optionally, redefine the pins used with the command: ChangeLCDpins RS, EN, D1, D2, D3, D4, RW
'- Define LCD size with the command:  InitLCDSize columns, lines
'- Setup the LCD with the command:    InitLCD
'- Display a line using the command:  PrintLCD LineNbr, Text$
'
' See documentation (LCD.PDF) for the schematic
' Maximite pin 11 is RS, pin 12 is EN
'    pins 13 to 16 are D4 to D7
' Duinomite pin 13 is RS, pin 14 is EN
'    pins 15 to 18 are D4 to D7 
'   (note: check the Duinomite LCD.pdf file - Geoff was considering pins 13 to 16 
'          are D4 to D7, and pin 17 = RS, pin 18 = EN ).
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' following should go at the beginning of code
' ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

' The following block of variables should not change once their values are set,
' so they are prefixed with a C_ to indicate that they are constants.

DIM C_LineStart(4)     ' an array to hold the start addresses for each line
C_Lines = 0            ' variable to hold the number of lines on the LCD
C_LineLength = 0       ' variable to hold the line length of the LCD
C_On  = 1              ' the following variables define parameter values for the subs
C_Off = 0
C_Cursor_Move   = 0
C_Display_Shift = 1
C_Decrement = 0
C_Increment = 1
C_Left  = 0
C_Right = 1
C_Normal = 0
C_Large  = 1
C_Cursor_Move   = 0
C_Display_Shift = 1
'
' The following block of variables are for internal use by the subroutines,
' mainly to supply missing information when the user only changes one attribute.
' For instance, to turn Blink on, you normally have to supply 2 other parameters.

' We STRONGLY suggest that users only use subroutines to change these variables.
' You could change them manually, but there's no guarantee they will work and, importantly,
' your code may fail with future versions of the library as more functionality is added to
' the subs.

_LCD_Display    = C_On
_LCD_Cursor     = C_On
_LCD_Blink      = C_On
_LCD_Type       = C_Cursor_Move
_LCD_MoveDir    = C_Increment
_LCD_Shift      = C_On
_LCD_Move_Type  = C_Cursor_Move
_LCD_Direction  = C_Right
_LCD_Bits       = 4
_LCD_Lines      = 2
_LCD_Font       = C_Normal
_LCD_RWenabled  = 0  ' variable indicating whether the Read Write pin is enabled (1=True) 

' Allocate names to pin numbers based on LCD.PDF and the comments above.
' The pin numbers must be changed using the ChangeLCDPins sub.

RS = 11
EN = 12
D1 = 13
D2 = 14
D3 = 15
D4 = 16
'
'
' This is where to put the subroutine calls to set up the library and LCD hardware.
' `````````````````````````````````````````````````````````````````````````````````

' **** Change the parameters to show whether you have a 16 x 2, or a 20 x 4 LCD screen.
InitLCDSize(20, 4)

' if you need to change pin numbers, do it before calling InitLCD
' eg ChangeLCDPins(13, 14, 15, 16, 17, 18)   ' - don't use unless you need to

InitLCD 


TestSubs

'--------------------------------------

SUB TestSubs
' Code to test subs
'    TestLCDClearScreen 	' ok
'    TestLCDHome 			' ok
'    TestLCDEntryMode		' ok
'    TestLCDControl			' ok
'    TestLCDShift			' ok
    TestLCDFunctionSet		' ok
'    TestLCDSetCGRAMaddr
'    TestLCDSetDDRAMaddr	' ok
'    TestLCDreadBusyFlag
    LCDClearScreen
    PrintLCD(1, "!!END OF TESTS!!")

End Sub


SUB TestLCDClearScreen

    PrintLCD(1,"Screen Clearing")
    PrintLCD(2,"in 2 seconds")
    Pause 2000
    LCDClearScreen
    Pause 2000
    PrintLCD(1,"Screen should")
    PrintLCD(2,"have cleared")
    WaitForUser
    LCDClearScreen

END SUB


SUB TestLCDHome

    LCD_Cursor(C_Off)
    LCD_Blink(C_On)
    PrintLCD(1,"Cursor should")
    PrintLCD(2,"move to Home")
    Pause 2000
    LCDHome
    ? "Cursor is OFF, Blink is ON, blinking insert point should be at Home position"
    WaitForUser
    LCDClearScreen

END SUB


SUB TestLCDEntryMode
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Sets cursor move direction and specifies display shift.
' These operations are performed during data write and read.
' argument #1 is the move direction (C_Increment or C_Decrement)
' argument #2 is the display shift (C_On or C_Off)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
	LOCAL StartCol, i, Msg$
	
	StartCol = C_LineLength \ 2		' start half way across the display
	
    LCDClearScreen
    LCDEntryMode(C_Increment, C_Off)
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	LCDEntryMode(C_Increment, C_On)
	? "Cursor Move direction = Increment, Display Shift = ON"
	? "Display shifts to the Left as the message on line 2 appears.
	? "Message is to the RIGHT of column StartCol."
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser

    LCDClearScreen
    LCDEntryMode(C_Increment, C_Off)
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	? "Cursor Move direction = Increment, Display Shift = OFF"
	? "Display stays still as the message on line 2 appears.
	? "Message is to the RIGHT of column StartCol."
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser    

    LCDClearScreen
    LCDEntryMode(C_Increment, C_Off)
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
	LCDEntryMode(C_Decrement, C_On)
	? "Cursor Move direction = Decrement, Display Shift = ON"
	? "Display shifts to the right as the message on line 2 appears.
	? "Reversed message is to the LEFT of column StartCol."
    Msg$ = "A test message"
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser

    LCDClearScreen
    LCDEntryMode(C_Increment, C_Off)
    PrintLCD(1,"Entry Mode Test")
	LCDLocate(2,StartCol)
    LCDEntryMode(C_Decrement, C_Off)
    Msg$ = "A test message"
	? "Cursor Move direction = Decrement, Display Shift = OFF"
	? "Display stays still as the message on line 2 appears.
	? "Reversed message is to the LEFT of column StartCol."
    FOR i = 1 TO LEN(Msg$)
        LCD_WriteByte ASC(MID$(Msg$, i, 1)), 1
        PAUSE 500
    NEXT i
	WaitForUser			

END SUB


SUB TestLCDControl
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set LCD display, cursor, and cursor blink controls
' argument #1 is the Display      status (C_On or C_Off)
' argument #2 is the Cursor       status (C_On or C_Off)
' argument #3 is the Cursor Blink status (C_On or C_Off)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDControl(Display, Cursor, Blink)

    PrintLCD 1, "  Hello  World"
    IF C_Lines > 1 THEN PrintLCD 2, "  Maximite LCD"
    IF C_Lines > 2 THEN PrintLCD 3, "with a 20 x 4 LCD,"
    IF C_Lines > 3 THEN PrintLCD 4, "using Subroutines."
    PAUSE 2000

    ' Tester should see the screen clear, and a blinking cursor
    '
    PrintLCD(1, "CLR SCR,Disp on")
    PrintLCD(2,"Cur on,Cur blink")
    ? "The screen should clear, the cursor should be on, and blinking"
    pause 2000

    lcdClearScreen
    PrintLCD(1, "Screen clear")
    WaitForUser

    LCDControl(C_On, C_On, C_On)    ' Display on, Cursor on, Cursor blink on
    PrintLCD(1, "Cur Blink ON")
    ? "The cursor should be on, and blinking"
    WaitForUser

    LCDControl(C_On, C_On, C_Off)
    PrintLCD(2, "Blink OFF")
    ? "The Cursor should be ON, but not blinking"
    WaitForUser
    lcdClearScreen

    LCDControl(C_On, C_Off, C_On)
    PrintLCD(1, "Cur OFF,Blink ON")
    ? "The Cursor should be OFF, and Blinking ON"
    WaitForUser

    LCDControl(C_On, C_Off, C_Off)
    PrintLCD(2, "Cur OFF,Blnk OFF")
    ? "The Cursor should be OFF, and not blinking"
    WaitForUser
    lcdClearScreen

    PrintLCD(1, "Display will OFF")
    Pause 2000
    LCDControl(C_Off, C_On, C_On)
    ? "The Display should be OFF, with a blinking Cursor (not visible, as display is off)"
    WaitForUser

    LCDControl(C_On, C_On, C_On)
    PrintLCD(2, "Display back ON")
    ? "The Display is back ON, with a blinking cursor"
    WaitForUser
    lcdClearScreen

END SUB


SUB TestLCDShift

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Cursor and Display shift.
' Moves cursor and shifts display without changing DDRAM contents.
' argument #1 is the type (C_Display_Shift or C_Cursor_Move)
' argument #2 is the shift direction (C_Left or C_Right)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDShift(Type, Direction)

    Local TestMsg$

    PrintLCD(1, "Cursor & Display")

' Load a 26 char message into Line 2 (noting that on a 16 x 2 character display
' we only see the first 16 characters, a 4 line display will wrap the line 2 text
' to line 4.)
' ------------------------------------------------------------------------------
    TestMsg$ = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    PrintLCD(2, TestMsg$)
    Pause 2000
'
' Now shift the message so the balance of the second line can be read
' -------------------------------------------------------------------
'
    ? "All lines will be Shifted Left for 40 characters"
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 40
        LCDShift(C_Display_Shift, C_Left)
        pause 500
    next _LCD
    ? "End of test for Display Shift Left. Cursor should be where it started."
    WaitForUser
'
' Note that this shifts both lines left, and you have to shift left 40 times
' to get back where you started from.
' Also notice how the text may move to different lines as the message is shifted.
'

    ? "All lines will be Shifted Right for 40 characters"
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 40
        LCDShift(C_Display_Shift, C_Right)
        pause 500
    next _LCD
    ? "End of test for Display Shift Right. Cursor should be where it started."
    WaitForUser

    ? "Cursor will be Shifted Left for 39 characters.  Display won't move."
    WaitForUser
    LCDLocate(2,40)	' end of line 2, or on 4 line display will be end of line 4
    for _LCD = 1 to 39 ' 39 not 40, because 'A' already in posn 1. 
        LCDShift(C_Cursor_Move, C_Left)
        pause 500
    next _LCD
    ? "End of test for Cursor Shift Left. Cursor should be where it started."
    WaitForUser

    ? "Cursor will be Shifted Right for 39 characters.  Display won't move."
    WaitForUser
    LCDLocate(2,1)
    for _LCD = 1 to 39
        LCDShift(C_Cursor_Move, C_Right)
        pause 500
    next _LCD
    ? "End of test for Cursor Shift Right. Cursor should be at end of line."
    WaitForUser

END SUB


SUB TestLCDFunctionSet
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the interface data length, number of display lines and
' character font.
' argument #1 is the number of bits used by the interface (4 or 8)
' note: this library is currently only set up to use 4 bits
' argument #2 is the number of lines (1 or 2) (Use 2 for 4 line)
' argument #3 is the font size (C_Large or C_Normal)
' Large is 5 x 10 dots, Normal is 5 x 8 dots.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDFunctionSet(_Bits, _Lines, _Font)

    ? "Note that normal text doesn't look any different when displayed"
    ? "with a large font.  If a large font is set, the first character" 
    ? "displayed on the LCD will be &hEA (a long j) which extends down"
    ? "below the normal character height."
    ?
    WaitForUser

    ' Bug: Can't figure out why the PrintLCD call (below) causes inverse
    ' squares to be displayed, and text not displayed.  Works OK if the
    ' LCDFunctionSet call is remarked out though. 

    LCDClearScreen
    ? "Setting LCD to 4 bit, 1 line, Normal sized font"    
    LCDFunctionSet(4, 1, C_Normal)
    PrintLCD(1, "4Bit 1Line Norml")
    WaitForUser
    
    LCDClearScreen
    ? "Setting LCD to 4 bit, 1 line, LARGE sized font"    
    LCDFunctionSet(4, 1, C_Large)
    PrintLCD(1, CHR$(&HEA) + "4Bit 1Line LRGE")
    WaitForUser
    
    LCDClearScreen
    ? "Setting LCD to 4 bit, TWO line, normal sized font"    
    LCDFunctionSet(4, 2, C_Normal)
    PrintLCD(1, "4Bit 2Line Norml")
    PrintLCD(2, "4Bit 2Line Norml")
    WaitForUser
    
    LCDClearScreen
    ? "Setting LCD to 4 bit, two line, LARGE sized font"    
    LCDFunctionSet(4, 2, C_Large)
    PrintLCD(1, CHR$(&HEA) + "4Bit 2Line LRGE")
    PrintLCD(2, CHR$(&HEA) + "4Bit 2Line LRGE")
    WaitForUser 
    
    LCDClearScreen
    ? "Setting LCD to 8 bit, one line, normal sized font"      
    LCDFunctionSet(8, 1, C_Normal)
    PrintLCD(1, "8Bit 1Line Norml")
    WaitForUser 
      
    LCDClearScreen
    ? "Setting LCD to 8 bit, one line, LARGE sized font"    
    LCDFunctionSet(8, 1, C_Large)
    PrintLCD(1, CHR$(&HEA) + "8Bit 1Line LRGE")
    WaitForUser 

    LCDClearScreen
    ? "Setting LCD to 8 bit, TWO line, normal sized font"    
    LCDFunctionSet(8, 2, C_Normal)
    PrintLCD(1, "8Bit 2Line Norml")
    PrintLCD(2, "8Bit 2Line Norml")
    WaitForUser 

    LCDClearScreen
    ? "Setting LCD to 8 bit, two line, LARGE sized font"    
    LCDFunctionSet(8, 2, C_Large)
    PrintLCD(1, CHR$(&HEA) + "8Bit 2Line LRGE")
    PrintLCD(2, CHR$(&HEA) + "8Bit 2Line LRGE")
    WaitForUser 

END SUB


SUB TestLCDSetCGRAMaddr
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' CGRAM data is sent and received after this setting.
' argument #1 is the CGRAM address (6 bits, so 0 to 63)
'
' In the character generator RAM, the user can rewrite character
' patterns by program.
' For 5 x  8 dots, eight character patterns can be written, and
' for 5 x 10 dots, four  character patterns can be written.
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDSetCGRAMaddr(_Addr)
END SUB


SUB TestLCDSetDDRAMaddr
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Set the Character Generator RAM (CGRAM) address.
' DDRAM data is sent and received after this setting.
' argument #1 is the DDRAM address (7 bits, so 0 to 127)
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDSetDDRAMaddr(_Addr)

	LOCAL i, addr
	? "Test the Set DDRAM Addr sub by writing an X in the second column of each line"
	LCDClearScreen
	FOR i = 1 to C_Lines
		addr = C_LineStart(i) + 2 
		LCDSetDDRAMaddr(addr)
		LCD_WriteByte ASC("X"), 1
	NEXT i
	WaitForUser
	 
END SUB


SUB TestLCDreadBusyFlag
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Read busy flag.
' Returns the Busy Flag (BF) indicating internal operation is being
' performed and returns address counter contents.
' bf=1 - still internally operating. bf=0 - instructions acceptable
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Called using LCDreadBusyFlag(_Status, _Addr)
	LOCAL	Status, Addr 
	' don't call unnecessary subs because we want to return control
	' to the MM as soon as possible (hopefully while the LCD is still busy)
	
    LCD_Nibble_Write &B0000 : LCD_Nibble_Write &B0001 ' clearing display is slow
    LCDreadBusyFlag(Status, Addr)
    IF Status = 1 then
        ? "Status is BUSY and address is "; Addr
    ELSE
        ? "Status is NOT BUSY and address is "; Addr
    ENDIF
    
END SUB

'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Wait for the user to press any key on the keyboard
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
SUB WaitForUser
	LOCAL Key$
	
	? "Press any key to continue " ;
	Do
		Key$ = INKEY$
	LOOP UNTIL Key$ <> ""
	?
	
END SUB

' and here's a test of one of the subroutines...
' TestMsg$ = "This is a really long message that we will scroll across the second line of the LCD screen."
' ScrollLCDMessage(2, TestMsg$)


' *** Example - Print a non-standard char. from the char. set held on the LCD
' ---------------------------------------------------------------------------
'_LCD_byte = &HF4 : pin(11) = 1 : SendByte ' print the ohm symbol
' ' note that the cursor didn't need to be moved
